from typing import List, Callable

import numpy as np


def get_reverse_modulus(number: int, mod: int = 26) -> int:
    result = 1

    while result % number != 0:
        result += mod

    return result // number


def modulo(number: int, mod: int = 26) -> int:
    while number < 0:
        number += mod

    return number % mod


def det(matrix: np.ndarray) -> int:
    return (matrix[0][0] * matrix[1][1]) - (matrix[0][1] * matrix[1][0])


def dot_product_with_multiple_matrices(matrices: List[np.ndarray]) -> np.ndarray:
    return np.linalg.multi_dot(matrices)


def create_nested_list_from_flat_list(flat_list: List[int]) -> List[List[int]]:
    return [[i] for i in flat_list]


def inverse_matrix(matrix: np.ndarray) -> List:
    return [[matrix[1][1], -1 * matrix[0][1]], [-1 * matrix[1][0], matrix[0][0]]]


def get_inverse_matrix(matrix: np.ndarray) -> List:
    det_of_matrix = det(matrix)

    inversed_matrix = inverse_matrix(matrix)

    if det_of_matrix == 1:
        return inversed_matrix

    multiplied_matrices = np.multiply(
        inversed_matrix, get_reverse_modulus(det_of_matrix)
    )

    return multiplied_matrices.tolist()


def apply_function_and_flatten_matrices(
    a: np.ndarray, b: np.ndarray, function: Callable
) -> List[int]:
    flat_matrix = function(a, b)

    return [flat_matrix[0, 0], flat_matrix[1, 1]]


def add_matrices_and_flatten(a: np.ndarray, b: np.ndarray) -> List[int]:
    return apply_function_and_flatten_matrices(a, b, np.add)


def sub_matrices(a: np.ndarray, b: np.ndarray) -> np.ndarray:
    return a - b



import numpy as np

from crypton.utils import math
from typing import List


def convert_int_to_space_index(index: int, space: str) -> str:
    return space[index]


def convert_space_index_to_int(character: str, space: str) -> int:
    return space.index(character)


def convert_ints_to_char(array: List, space: str) -> List:
    space_length = len(space)
    return [
        convert_int_to_space_index(math.modulo(i, space_length), space) for i in array
    ]


def convert_nested_ints_to_char(array: np.ndarray, space: str) -> List:
    space_length = len(space)
    return [
        "".join(
            [convert_int_to_space_index(math.modulo(j, space_length), space) for j in i]
        )
        for i in array
    ]



from crypton.utils.math import modulo
from functools import partial
from crypton.utils.space import convert_space_index_to_int
from typing import Callable, List


def apply_func_to_char(c: int, f: Callable, mod: int = 26) -> int:
    result = f(c)

    return modulo(result, mod)


def encrypt_decrypt_helper(
    plain_text: str, a: int, b: int, f: Callable, space: str
) -> str:

    encrypt_func = partial(apply_func_to_char, f=f(a, b), mod=len(space))

    text_as_list_of_ints = list(
        map(partial(convert_space_index_to_int, space=space), plain_text)
    )

    result_as_list_of_ints = list(map(encrypt_func, text_as_list_of_ints))

    result_as_list_of_strs = list(map(lambda x: space[x], result_as_list_of_ints))

    return "".join(result_as_list_of_strs)


def get_pairs_of_int_two_from_text(text: str, space: str) -> List[List[int]]:
    return [
        [space.index(text[i]), space.index(text[i + 1])]
        for i in range(0, len(text) - 1, 2)
    ]
